__description__ = "a module that houses methods used to produce visualizations"
__author__ = "Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import os
from collections import defaultdict
import numpy as np

import matplotlib
matplotlib.use('Agg')
from matplotlib import figure
from matplotlib import pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

from idq import features

#-------------------------------------------------
### default plot settings

matplotlib.rcParams.update({
    "font.size": 9.0,
    "axes.titlesize": 10.0,
    "axes.labelsize": 10.0,
    "xtick.labelsize": 8.0,
    "ytick.labelsize": 8.0,
    "legend.fontsize": 8.0,
    "figure.dpi": 300,
    "savefig.dpi": 300,
    "text.usetex": True,
    "path.simplify": True,
    "font.family": 'serif',
})

DEFAULT_COLORMAP = 'copper_r'

#-------------------------------------------------
### plotting functions

def roc_curve(eff, fap, nickname='', xscale='linear', **kwargs):
    """
    Plots a ROC curve based on eff and FAP timeseries generated from a
    classifier, as well as any extra keyword args for plot settings.
    """
    figs, axes = plt.subplots()

    axes.set_xscale(xscale)
    axes.set_yscale('linear')

    random_guess_line = np.linspace(1e-5, 1)

    axes.plot(fap, eff, color='darkorange')
    axes.plot(random_guess_line, random_guess_line, color='navy', linestyle='--')

    axes.set_title(r"ROC Curve for %s" % nickname.replace('_', ' '))
    axes.set_xlabel(r"False Alarm Probability")
    axes.set_ylabel(r"Glitch Detection Efficiency")

    return figs, axes

def timeseries(times, timeseries, nickname='', timeseries_name='', **kwargs):
    """
    Plots a timeseries generated from a trained classifier,
    as well as any extra keyword args for plot settings.
    """
    gps_end = np.ceil(times[-1]) ### round up to integer GPS for aesthetics
    times = times - times[-1]    ### reference time from gps end

    figs, axes = plt.subplots()

    axes.set_xscale('linear')
    axes.set_yscale('linear')

    axes.plot(times, timeseries, color='darkorange')

    axes.set_title(r"%s timeseries for %s" % (timeseries_name, nickname.replace('_', ' ')))
    axes.set_xlabel(r"Time [s] from %f"%gps_end )
    axes.set_ylabel(r"%s" % timeseries_name)

    return figs, axes

def latency_percentile(latencies, stage='', **kwargs):
    """
    Plots a set of latency measurements in a percentile-type plot,
    as well as any extra keyword args for plot settings.
    """
    ### histogram latency measurements and compute cdf
    hist, bin_edges = np.histogram(latencies, bins=50, density=True)
    spacing = bin_edges[1] - bin_edges[0]
    cdf = np.cumsum(hist) * spacing
    binned_latencies = bin_edges[1:]

    figs, axes = plt.subplots()

    axes.set_xscale('log')
    axes.set_yscale('linear')

    axes.plot(1 - cdf, binned_latencies, color='darkorange')

    axes.set_title(r"Latency by Percentile Distribution at %s"%stage)
    axes.set_xlabel(r"Percentile")
    axes.set_ylabel(r"Latency")

    ### set x-axis to display percentiles
    axes.set_xlim(axes.get_xlim()[::-1]) ### reverse x-axis
    plt.xticks(*percentile_ticks(cdf[-2])) ### grab second-to-last point for upper bound to not cause numerical issues

    return figs, axes

#-------------------------------------------------
### DQR plots

PGLITCH_LABEL = '$p(\mathrm{glitch}|\mathrm{aux})$'
LOGLIKE_LABEL = '$\ln \Lambda^\mathrm{glitch}_\mathrm{clean}$'
EFF_FAP_LABEL = 'Efficiency/FAP'
FAP_LABEL = 'FAP'
RANK_LABEL = 'Rank'

CUM_GLITCH_LIKE_LABEL = '$P(\mathrm{aux}|\mathrm{glitch})$'
GLITCH_LIKE_LABEL = '$p(\mathrm{aux}|\mathrm{glitch})$'
CUM_CLEAN_LIKE_LABEL = '$P(\mathrm{aux}|\mathrm{clean})$'
CLEAN_LIKE_LABEL = '$p(\mathrm{aux}|\mathrm{clean})$'

def dqr_timeseries(nicknames, series, start, end, t0, fig=None, gch_gps=[]):
    """plot the series information in a standard way
    """
    if fig is None:
        fig = plt.figure()
    ax_pglitch = plt.subplot(5, 1, 1)
    ax_loglike = plt.subplot(5, 1, 2)
    ax_eff_fap = plt.subplot(5, 1, 3)
    ax_fap = plt.subplot(5, 1, 4)
    ax_rank = plt.subplot(5, 1, 5)

    # plot
    for nickname in nicknames:
        for s in series[nickname]:
            if 'times' in s['data']:
                times = s['data']['times'] - t0
            else:
                times = np.arange(s['t0'], s['deltaT']*len(s['data'].values()[0]), dt) - t0

            for key, val in s['data'].items():
                ### NOTE: the way we extract channels might be a bit fragile...
                if 'LOGLIKE' in key:
                    ax_loglike.plot(times, val, alpha=0.75, label=nickname)
                elif 'PGLITCH' in key:
                    ax_pglitch.plot(times, val, alpha=0.75, label=nickname)
                elif 'RANK' in key:
                    ax_rank.plot(times, val, alpha=0.75, label=nickname)
                elif 'EFF' in key:
                    eff_key = key
                elif 'FAP' in key:
                    fap_key = key
                    ax_fap.plot(times, val, alpha=0.75, label=nickname)

            # plot eff/fap as a special case
            ax_eff_fap.plot(times, s['data'][eff_key]/s['data'][fap_key], alpha=0.75, label=nickname)

    # decorate
    ax_pglitch.set_ylabel(PGLITCH_LABEL)
    ax_loglike.set_ylabel(LOGLIKE_LABEL)
    ax_eff_fap.set_ylabel(EFF_FAP_LABEL)
    ax_fap.set_ylabel(FAP_LABEL)
    ax_rank.set_ylabel(RANK_LABEL)

    ax_rank.set_xlabel('Time [sec] relative to %.3f'%(t0))

    ax_eff_fap.set_yscale('log')
    ax_fap.set_yscale('log')

    xlim = (start-t0, end-t0)
    ax_rank.set_xlim(xlim)
    for ax in [ax_pglitch, ax_loglike, ax_eff_fap, ax_fap]:
        plt.setp(ax.get_xticklabels(), visible=False)
        ax.set_xlim(xlim)

    ax_rank.legend(loc='best')

    # add annotation gps
    if len(gch_gps):
        gch_gps = np.array(gch_gps)
        gch_gps = gch_gps[(start<=gch_gps)*(gch_gps<end)] # filter by only the times that will show up
        gch_gps = gch_gps-t0
        verts = np.empty((2, len(gch_gps)), dtype=float)
        gch_gps = [gch_gps]*2
        for ax in [ax_pglitch, ax_loglike, ax_eff_fap, ax_fap, ax_rank]:
            ymin, ymax = ax.get_ylim()
            verts[0,:] = ymin
            verts[1,:] = ymax
            ax.plot(gch_gps, verts, color='grey', linestyle='dashed', alpha=0.10)
            ax.set_ylim(ymin=ymin, ymax=ymax)

    plt.subplots_adjust(hspace=0.10, wspace=0.10)
    return fig

def dqr_roc(nicknames, series, quivers, calibmaps, fig=None):
    """plot the ROC curves in a standard way
    """
    if fig is None:
        fig = plt.figure()
    ax = fig.gca()

    xmin = 1e-3
    for nickname in nicknames:
        quiver = quivers[nickname]
        ### actually plot data
        xmin = min(xmin, 0.1/len(quiver))

        ### directly from ranks in evaluated quivers
        color = _quiver2roc(ax, quiver, label=nickname+": "+RANK_LABEL, linestyle='dashed', alpha=0.25, marker='.')

        ### plot based on p(glitch) assigned by timeseries
        # update vectors to reflect the p(glitch) they were assigned
        for key in series[nickname][0]['data'].keys(): ### find the relevant key
            if 'PGLITCH' in key:
                break
        q = features.Quiver() ### make a copy so we don't muck up the original data
        for v in quiver:
            v = features.FeatureVector(v.gps, v.label, None)
            v.rank = _series2val(series[nickname], v.gps, key)
            q.append(v)
        _quiver2roc(ax, q, color=color, label=nickname+": "+PGLITCH_LABEL, linestyle='solid', alpha=0.25, marker=None)

        ### plot calibration maps
        ranks = np.array(sorted(set(quiver.ranks)))
        for calibmap in calibmaps[nickname].values():
            for r in ranks:
                fap = 1.-calibmap.cln_cdf_quantile([r], [0.16, 0.84])[0]
                eff = 1.-calibmap.gch_cdf_quantile([r], [0.16, 0.84])[0]
                ax.fill_between(fap, [eff[0]]*2, [eff[1]]*2, color=color, alpha=0.15)
            ax.plot(*calibmap.roc(ranks=ranks), marker='s', color=color, linestyle='-.', drawstyle='steps', label=nickname+': [%s, %s)'%tuple(calibmap.hash.split('_')))

    # plot a fair coin's flip
    line = np.logspace(np.log10(xmin), 0, 1001)
    ax.plot(line, line, color='k', linestyle='dashed', alpha=0.25)

    ### decorate
    ax.set_xlabel('FAP')
    ax.set_ylabel('Efficiency')

    ax.set_xscale('log')

    ax.set_ylim(ymin=-0.01, ymax=1.01)
    ax.set_xlim(xmin=xmin, xmax=1.01)

    ax.legend(loc='lower right')

    plt.subplots_adjust(hspace=0.10, wspace=0.10)
    return fig

def _series2val(series, gps, key):
    for s in series:
        t = s['data']['times'] ### expect that this has been added at this point
        dt = s['deltaT']
        if (t[0]<=gps) and (gps<t[-1]+dt):
            return np.interp(gps, t, s['data'][key])
    return np.nan ### could not find anything         

def _quiver2roc(ax, quiver, color=None, label=None, linestyle='solid', alpha=0.25, marker='o'):
    """helper function for generating an ROC curve from a quiver
    """
    ranks = np.array(sorted(set(quiver.ranks))) ### the list of all ranks observed in quivers
    gch, cln = quiver.vectors2classes()
    
    ### actually plot data
    ### directly from ranks in evaluated quivers
    eff = np.zeros_like(ranks, dtype=float)
    fap = np.zeros_like(ranks, dtype=float)
    
    for r in gch.ranks:
        eff += r >= ranks
    eff /= len(gch)
    seff = (eff*(1-eff)/len(gch))**0.5

    for r in cln.ranks:
        fap += r > ranks
    fap /= len(cln)
    sfap = (fap*(1-fap)/len(cln))**0.5

    # plot uncertainty regions
    if color is not None:
        ax.plot(fap, eff, drawstyle='steps', color=color, marker=marker, label=label, linestyle=linestyle)
    else:
        color = ax.plot(fap, eff, drawstyle='steps', marker=marker, label=label, linestyle=linestyle)[0].get_color()
    for e, se, f, sf, r in zip(eff, seff, fap, sfap, ranks):
        ax.fill_between([f-sf, f+sf], [e-se]*2, [e+se]*2, color=color, alpha=alpha)

    return color

def dqr_calibration_distribs(nicknames, quivers, calibmaps, fig=None):
    """plot the distributions over observed ranks and the calibration KDEs
    """
    if fig is None:
        fig = plt.figure()
    ax_cum_gch = plt.subplot(2,2,1)
    ax_cum_cln = plt.subplot(2,2,2)
    ax_gch = plt.subplot(2,2,3)
    ax_cln = plt.subplot(2,2,4)

    # plot histograms of ranks within quiver
    for nickname in nicknames:
#        gch, cln = quivers[nickname].vectors2classes()
#        for ax_cum, ax, q in [(ax_cum_gch, ax_gch, gch), (ax_cum_cln, ax_cln, cln)]:
#            ax_cum.hist(q.ranks, histtype='step', normed=True, label='Quiver', cumulative=-1)
#            ax.hist(q.ranks, histtype='step', normed=True, label='Quiver')

        # plot distirbutions stored within calibration maps
        for calibmap in calibmaps[nickname].values():
            label = nickname+': [%s, %s)'%tuple(calibmap.hash.split('_'))
            for ax_cum, ax, kde in [(ax_cum_gch, ax_gch, calibmap._gch_kde), (ax_cum_cln, ax_cln, calibmap._cln_kde)]:
                # control the number and placement of bins by hand so we don't end up with enormous things
                num_bins = min(100, max(10, 2*len(kde.obs)**0.5))
                bins = np.linspace(0, 1, num_bins+1)

                # cumulative distributions
                cdf = 1.-kde.cdf(kde._interp_x)
                low = 1.-kde.cdf_quantile(kde._interp_x, 0.16)
                high = 1.-kde.cdf_quantile(kde._interp_x, 0.84)
                color = ax_cum.plot(kde._interp_x, cdf, label=label, alpha=0.75)[0].get_color()
                ax_cum.fill_between(kde._interp_x, low, high, color=color, alpha=0.25)
                ax_cum.hist(kde.obs, bins=bins, color=color, histtype='step', normed=True, cumulative=-1, linestyle='dashed', alpha=0.75)

                # differential distributions
                pdf = kde.pdf(kde._interp_x)
                low = kde.pdf_quantile(kde._interp_x, 0.16)
                high = kde.pdf_quantile(kde._interp_x, 0.84)
                ax.plot(kde._interp_x, pdf, color=color, label=label, alpha=0.75)
                ax.fill_between(kde._interp_x, low, high, color=color, alpha=0.25)
                ax.hist(kde.obs, bins=bins, color=color, histtype='step', normed=True, label=label, linestyle='dashed', alpha=0.75)

    ### decorate
    ax_cum_gch.set_ylabel(CUM_GLITCH_LIKE_LABEL)
    ax_cum_cln.set_ylabel(CUM_CLEAN_LIKE_LABEL)
    ax_gch.set_ylabel(GLITCH_LIKE_LABEL)
    ax_cln.set_ylabel(CLEAN_LIKE_LABEL)

    xlim = (0., 1.)
    ylim = (0., 1.)
    for ax in [ax_cum_gch, ax_cum_cln]:
        plt.setp(ax.get_xticklabels(), visible=False)
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

    for ax in [ax_gch, ax_cln]:
        ax.set_xlabel(RANK_LABEL)
        ax.set_xlim(xlim)

    for ax in [ax_cum_cln, ax_cln]:
        ax.yaxis.tick_right()
        ax.yaxis.set_label_position('right')

    ax_cum_gch.legend(loc='best')

    plt.subplots_adjust(hspace=0.05, wspace=0.05)
    return fig    

def dqr_calibration_coverage(nicknames, series, quivers, fig=None):
    """plot the calibration coverage in a standard way
    """
    if fig is None:
        fig = plt.figure()
    ax_eff = plt.subplot(1,2,1)
    ax_fap = plt.subplot(1,2,2)

    # plot reference lines
    x = np.linspace(0, 1, 1001)
    ax_eff.plot(x, x, color='k', alpha=0.5, linestyle='dashed')
    ax_fap.plot(x, x, color='k', alpha=0.5, linestyle='dashed')

    ### prepare data
    for nickname in nicknames:
        for key in series[nickname][0]['data'].keys(): ### find the relevant key
            if 'EFF' in key:
                effkey = key
            elif 'FAP' in key:
                fapkey = key

        gch, cln = quivers[nickname].vectors2classes()

        effs = np.zeros_like(x, dtype=float)
        for v in gch:
            effs += _series2val(series[nickname], v.gps, effkey) <= x
        effs /= len(gch)
        seffs = (effs*(1-effs)/len(gch))**0.5

        faps = np.zeros_like(x, dtype=float)
        for v in cln:
            faps += _series2val(series[nickname], v.gps, fapkey) <= x 
        faps /= len(cln)
        sfaps = (faps*(1-faps)/len(cln))**0.5

        # plot
        color = ax_eff.plot(x, effs, label=nickname)[0].get_color()
        ax_eff.fill_between(x, effs-seffs, effs+seffs, color=color, alpha=0.25)

        ax_fap.plot(x, faps, color=color, label=nickname)
        ax_fap.fill_between(x, faps-sfaps, faps+sfaps, color=color, alpha=0.25)

    ax_fap.legend(loc='best')

    # decorate
    ax_eff.set_xlabel('Nominal Efficiency')
    ax_eff.set_ylabel('Observed Fraction')

    ax_fap.set_xlabel('Nominal FAP')
    ax_fap.set_ylabel('Observed Fraction')
    ax_fap.yaxis.tick_right()
    ax_fap.yaxis.set_label_position('right')

    lim = (-0.01, 1.01)
    for ax in [ax_eff, ax_fap]:
        ax.set_xlim(lim)
        ax.set_ylim(lim)

    plt.subplots_adjust(hspace=0.05, wspace=0.05)
    return fig

def dqr_featureimportance(nicknames, models, classifiers, quiverfactory, start, end, t0):
    """delegate to models heavily here
    """
    figs = dict()
    tabs = dict()
    for nickname in nicknames:
        f = dict()
        t = dict()
        classifier = classifiers[nickname]
        for key, model in models[nickname].items():
            classifier.model = model ### assign this so internal plumbing works out
            ### make a figure
            try:
                f[key] = classifier.feature_importance_figure(quiverfactory, start, end, t0)
            except NotImplementedError:
                f[key] = None
            ### make a table
            try:
                t[key] = classifier.feature_importance_table(quiverfactory, start, end, t0)
            except NotImplementedError:
                t[key] = None
        figs[nickname] = f
        tabs[nickname] = t
    return figs, tabs

#-------------------------------------------------
### plotting utilities

def close(fig):
    """
    delegate to pyplot.close
    """
    plt.close(fig)

def percentile_ticks(cdf_upper_bound):
    """
    Given a CDF upper bound, generates all the percentiles
    needed to map to ticks in an associated plot.

    Returns two lists in the form (1 - percentile, percentile).
    """
    endpoint = np.floor(np.log10(1 - cdf_upper_bound))
    percentile = 1 - np.logspace(0, endpoint, num=(-endpoint+1))

    return 1 - percentile, percentile
